/*
//
// BEGIN SONGBIRD GPL
//
// This file is part of the Songbird web player.
//
// Copyright(c) 2005-2008 POTI, Inc.
// http://songbirdnest.com
//
// This file may be licensed under the terms of of the
// GNU General Public License Version 2 (the "GPL").
//
// Software distributed under the License is distributed
// on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
// express or implied. See the GPL for the specific language
// governing rights and limitations.
//
// You should have received a copy of the GPL along with this
// program. If not, go to http://www.gnu.org/licenses/gpl.html
// or write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//
// END SONGBIRD GPL
//
*/

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cu = Components.utils;
const Cr = Components.results;

const CLASS_ID = Components.ID("{2f1f0218-a3de-46ad-baf2-94217f6574c3}");
const CLASS_NAME = "Daap Library Manager Service";
const CONTRACT_ID = "@songbirdnest.com/daap/library-manager;1";
/* Extension ID */
const EXTENSION_UUID = "daapclient@songbirdnest.com";
const PROPERTY_FILE = "chrome://daapclient/locale/daapclient.properties";

if (typeof(XPCOMUtils) == "undefined") {
  Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
  if (!XPCOMUtils)
    throw new Error("Import of XPCOMUtils module failed!");
}


/* global variables -----------------------------------------------------------*/
var libraryManager = null;
var SPS = null;
var BMS = null;
var librarySPS = null;
var ioService = null;
var daapLibManager = null;
var commandsManager = null; 
var storageService = null;
var l10n = null;

var start = null;


const PlaylistCommandsBuilder = new Components.
            Constructor("@songbirdnest.com/Songbird/PlaylistCommandsBuilder;1", 
                        "sbIPlaylistCommandsBuilder");



/*=============================================================================*/
/* SbDaapLibManager                                                            */
/*=============================================================================*/

/* SbDaapLibManager constructor -----------------------------------------------*/
function SbDaapLibManager() 
{
  if (typeof(DProperty) == "undefined") {
    Components.utils.import("resource://daapclient/DaapProperties.jsm");
  }
  if (!(commandsManager instanceof Ci.sbIPlaylistCommandsManager)) {
    commandsManager = Cc["@songbirdnest.com/Songbird/PlaylistCommandsManager;1"]
                       .getService(Ci.sbIPlaylistCommandsManager);
  }

  this._libraries = {};
  this._static = {};
  this._commands = null;

}

/* SbDaapLibManager implementation ---------------------------------------------*/
SbDaapLibManager.prototype = {

  /* custom factory */
  _xpcom_factory: SbDaapLibManagerFactory,

  /* nsISupports ----------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapLibraryManager, 
                                          Ci.nsIClassInfo]),

  /* nsIClassInfo, so callers don't have to QI from JS -------------------*/
  classDescription        : CLASS_NAME,
  classID                 : CLASS_ID,
  contractID              : CONTRACT_ID,
  flags                   : Ci.nsIClassInfo.THREADSAFE,
  implementationLanguage  : Ci.nsIProgrammingLanguage.JAVASCRIPT,
  getHelperForLanguage    : function(aLanguage) { return null; },
  getInterfaces : function(count) {
    var interfaces = [Ci.sbIDaapLibraryManager, Ci.nsIClassInfo, Ci.nsISupports];
    count.value = interfaces.length;
    return interfaces;
  },


  /* private data ---------------------------------------------------------*/
  _libraries: null,
  _static: null,
  _commands: null,

  /* sbIDaapLibraryManager ------------------------------------------------*/


  /* void init(); */
  init: function SbDaapLibraryManager_init ()
  {
    this.restoreLibraries();
  },

  /* void shutdown(); */
  shutdown: function SbDaapLibraryManager_shutdown ()
  {
  },


  /* registerLibrary:
   *  Register a new library.
   *
   *  \param aDatabaseId    a unique database ID
   *  \param aMachineName   the name
   *  \param aAddress       daap server IP address 
   *  \param aPort          daap server TCP port
   *  \param aLoadAtStartup load the library at startup
   *
   *  \return sbILibrary    the related library
   *
   **/
 
  /* sbILibrary registerLibrary (in AString databaseId, 
                          in AString machineName, 
                          in AString address, 
                          in PRUInt16 port,
                          [optional] in boolean aLoadAtStartup);  */

  registerLibrary: function SbDaapLibManager_registerLibrary (aDatabaseId, 
                                                              aMachineName, 
                                                              aAddress, aPort,
                                                              aLoadAtStartup)
  {
    /* preconditions */
    if (!aDatabaseId || !aDatabaseId.length
        || !aMachineName || !aMachineName.length
        || !aAddress || !aAddress.length
        || !aPort ) {
      /* abort */
      throw Components.results.NS_ERROR_INVALID_ARG; 
    }

    if (this._libraries[aDatabaseId]) {
      /* already registered */
      return this.getLibrary(aDatabaseId);
    }

    /* register the library */
    this._libraries[aDatabaseId] = new DaapLibrary();
    this._libraries[aDatabaseId].load(aDatabaseId, aMachineName, aAddress, aPort);

    /* add library actions */
    this.addCommands(aDatabaseId);

    /* save the library for loading it at startup */
    if (aLoadAtStartup) {
      this.saveLibraryConfig(aDatabaseId, aMachineName, aAddress, aPort);
      this._static[aDatabaseId] = {id: aDatabaseId, name: aMachineName, 
                          address: aAddress, port: aPort};
    }


    return this.getLibrary(aDatabaseId);
  },  //registerLibrary



  /* unregisterLibrary:
   *  Unregister a library.
   *
   *  \param  aDatabaseId   the library ID
   *
   **/

  /* void unregisterLibrary (in AString aDatabaseId); */
  unregisterLibrary: function SbDaapLibManager_unregisterLibrary (aDatabaseId)
  {
    /* preconditions */
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      return;
    }

    /* add library actions */
    this.removeCommands(aDatabaseId);

    /* unregister the library */
    this._libraries[aDatabaseId].unload();
    this._libraries[aDatabaseId] = undefined;
      
    return;
  },  //unregisterLibrary


 
  /* void open (in sbIMediaList aMediaList); */
  open: function SbDaapLibManager_open (aMediaList)
  {
    if (!aMediaList) {
      /* abort */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    let databaseId = aMediaList.library.getProperty(DProperty.databaseId);
    if (!databaseId || !this._libraries[databaseId]) {
      /* not registered */
      return;
    }

    if (aMediaList instanceof Ci.sbILibrary) {
      /* it's a library */
      this._libraries[databaseId].open();

      /* after opening check if the name has changed */
      let library = this.getLibrary(databaseId);
      if (library instanceof Ci.sbILibrary
          && this._static[databaseId]
          && this._static[databaseId].name != library.name) {
        /* update the saved library name */
        this.saveLibraryConfig(databaseId, library.name, 
                                this._static[databaseId].address, 
                                this._static[databaseId].port);
      }

    }
    else {
      /* it's a playlist */
      this._libraries[databaseId].openPlaylist(aMediaList);
    }

    return;
  },  //open


  /* void delete (in AString aDatabaseId); */
  delete: function SbDaapLibManager_delete (aDatabaseId)
  {
    if (!aDatabaseId) {
      /* abort */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    if (!this.isStatic(aDatabaseId)) {
      /* not static or not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }

    if (this.isOpen(aDatabaseId)) {
      try {
        this.close(aDatabaseId);
      }
      catch (e) {}
    }

    let library = this.getLibrary(aDatabaseId);

    /* delete from start up */
    this.deleteFromConfig(aDatabaseId);
    /* unregister from us */
    this.unregisterLibrary(aDatabaseId);

    /* delete from disk */
    //database.delete();
    try {
      //XXX this never works... I must ask why on IRC
      let file = library.creationParameters
                        .getPropertyAsInterface("databaseFile", Ci.nsIFile);
      file.remove (false); //false stands for not recursive, ignored for files
    }
    catch (e) {
      Components.utils.reportError(e);
    }

    return;
  },  //delete


  /* getLibrary:
   *  Returns the related library.
   *
   *  \param aDatabaseId  the library ID
   *
   **/
    
  /* sbILibrary getLibrary (in AString aDatabaseId); */
  getLibrary: function SbDaapLibManager_getLibrary (aDatabaseId)
  {
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }

    return this._libraries[aDatabaseId].getLibrary();
  },  //getLibrary



  /* close:
   *  Close the library and logout from the server.
   *
   *  \param aDatabaseId  the library ID
   *
   **/

  /* void close (in AString databaseId); */
  close: function SbDaapLibManager_close (aDatabaseId)
  {
    /* preconditions */  
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }
    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }

    /* delegate to the Daap Library */
    this._libraries[aDatabaseId].close();

    return;
  }, //close



  /* isOpen:
   *  Determine if a library is open or not
   * 
   * \param databaseId  the library ID
   * \return boolean
   *
   **/

  /* boolean isOpen (in AString databaseId); */
  isOpen: function SbDaapLibManager_isOpen (aDatabaseId)
  {
    /* preconditions */  
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }
    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }

    /* delegate to the Daap Library */
    return this._libraries[aDatabaseId].isOpen();

  }, //isOpen



  /* isStatic:
   *  Determine if a library is static or not
   * 
   * \param databaseId  the library ID
   * \return boolean
   *
   **/

  /* boolean isStatic (in AString databaseId); */
  isStatic: function SbDaapLibManager_isStatic (aDatabaseId)
  {
    /* preconditions */  
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }
    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }

    /* delegate to the Daap Library */
    return typeof(this._static[aDatabaseId]) != "undefined";

  }, //isStatic



  /* forceUpdate:
   *  Setup an update for this library on next access.
   *
   * \param aDatabaseId the database ID
   *
   **/
  /* void forceUpdate(in AString databaseId); */

  forceUpdate: function SbDaapLibManager_forceUpdate (aDatabaseId)
  {
    /* preconditions */  
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }
    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }

    /* delegate to the Daap Library */
    return this._libraries[aDatabaseId].forceUpdate();

  },  //forceUpdate
 
  /* private functions ----------------------------------------------------*/


  /* addCommands:
   *  Register daap specific commands for a library: logout, force sync, ...
   *
   *  \param aDatabaseId   the library ID which the commands are registered for
   *
   **/

  addCommands: function SbDaapLibManager_addCommands (aDatabaseId)
  {
    /* preconditions */
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }


    let library = this.getLibrary(aDatabaseId);
    let commands = this.getCommands();


    /* register this command */
    commandsManager.registerPlaylistCommandsMediaList(library.guid, "", commands);

    return;
  },  //addCommands


  /* removeCommands:
   *  Unregister daap specific commands for a library: logout, force sync, ...
   *
   *  \param aDatabaseId   the library ID which the commands were registered for
   *
   **/

  removeCommands: function SbDaapLibManager_addCommands (aDatabaseId)
  {
    /* preconditions */
    if (!aDatabaseId) {
      /* invalid id */
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    if (!this._libraries[aDatabaseId]) {
      /* not registered */
      throw Components.results.NS_ERROR_FAILURE;
    }


    let library = this.getLibrary(aDatabaseId);
    let commands = this.getCommands();

    /* unregister this command */
    commandsManager.unregisterPlaylistCommandsMediaList(library.guid, "", commands);

    return;
  },  //removeCommands

  
  /* getCommands:
   *  Macke the library commands.
   *
   *  \return sbIPlaylistCommandsBuilder  the commands
   *
   **/
  getCommands: function SbDaapLibManager_getCommands ()
  {
    if (this._commands instanceof Ci.sbIPlaylistCommandsBuilder) {
      return this._commands;
    }
    if (!(l10n instanceof Ci.nsIStringBundleService)) {
      l10n = Cc["@mozilla.org/intl/stringbundle;1"]
              .getService(Ci.nsIStringBundleService)
              .createBundle(PROPERTY_FILE);
    }

    this._commands = new PlaylistCommandsBuilder();
    
    /* Create the commands */
    this._commands.appendAction(null,
                          "daap_cmd_force_update",            //name
                          l10n.GetStringFromName("CmdForceUpdate"), //label
                          l10n.GetStringFromName("CmdForceUpdateTooltip"), //tooltip
                          SbDaapLibManager_cmdForceUpdateExecute ); //callback
    this._commands.setCommandVisibleCallback(null,
                                        "daap_cmd_force_update", //command name
                                        SbDaapLibManager_cmdForceUpdateVisible ); //callback

    this._commands.appendAction(null,
                          "daap_cmd_logout",            //name
                          l10n.GetStringFromName("CmdLogout"),  //label
                          l10n.GetStringFromName("CmdLogoutTooltip"), //tooltip
                          SbDaapLibManager_cmdLogoutExecute ); //callback
    this._commands.setCommandVisibleCallback(null,
                                        "daap_cmd_logout", //command name
                                        SbDaapLibManager_cmdLogoutVisible ); //callback

    this._commands.appendAction(null,
                          "daap_cmd_delete",            //name
                          l10n.GetStringFromName("CmdDelete"),  //label
                          l10n.GetStringFromName("CmdDeleteTooltip"), //tooltip
                          SbDaapLibManager_cmdDeleteExecute ); //callback
    this._commands.setCommandVisibleCallback(null,
                                        "daap_cmd_delete", //command name
                                        SbDaapLibManager_cmdDeleteVisible ); //callback
    
    
    return this._commands;
  },//getCommands


  /* saveLibraryConfig:
   *  Save library paramaters on persistent storage.
   *
   * \param aDatabaseId
   * \param aMachineName
   * \param aAddress
   * \param aPort
   *
   **/

  saveLibraryConfig: function SbDaapLibManager_saveLibraryConfig (aDatabaseId, 
                                                                  aMachineName, 
                                                                  aAddress, 
                                                                  aPort)
  {
    /* preconditions */
    if (!aDatabaseId || !aDatabaseId.length
        || !aMachineName || !aMachineName.length
        || !aAddress || !aAddress.length
        || !aPort ) {
      /* abort */
      throw Components.results.NS_ERROR_INVALID_ARG; 
    }

    let dbConn = this.openConfigDatabase();


    let insertQuery = "REPLACE INTO \"main\".\"libraries\" "
                        +"(\"id\",\"name\",\"address\",\"port\") "
                        +"VALUES("
                        +"\"" + aDatabaseId + "\","
                        +"\"" + aMachineName + "\","
                        +"\"" + aAddress + "\","
                        + aPort
                        +");"

    dbConn.executeSimpleSQL(insertQuery);
    dbConn.close();
    
    return;
  }, //saveLibraryConfig


  /* restoreLibraries:
   *  Recover library paramaters from persistent storage
   *  and register them with us.
   *
   **/

  restoreLibraries: function SbDaapLibManager_restoreLibraries ()
  {
    let dbConn = this.openConfigDatabase();

    const selectQuery = "SELECT \"id\",\"name\",\"address\",\"port\" "
                        +"FROM \"main\".\"libraries\"; "; 
    let statement = dbConn.createStatement(selectQuery);

    while (statement.executeStep()) {
      let id = statement.getString(0);
      let name = statement.getString(1);
      let address = statement.getString(2);
      let port = statement.getInt32(3);

      try {
        //TODO: check this....
        let library = this.registerLibrary(id, name, address, port); //XXX: true??
        if (library instanceof Ci.sbILibrary) {
          this._static[id] = {id: id, name: name, address: address, port: port};
        }
      }
      catch (e) {
        Components.utils.reportError(e);
        Components.utils.reportError("[Daap Client] Cannot restore library: "
                                      +"id="+id+"; name="+name+";"
                                      +"address="+address+"; port="+port);
      }
        
    }

    statement.finalize();
    dbConn.close();

    return ;
  }, //restoreLibraries


  /* openConfigDatabase:
   *  Open the config database, it will be created if necesary.
   *
   * \return mozIStorageConnection the database conenction
   *
   **/
  openConfigDatabase: function SbDaapLibManager_openConfigDatabase ()
  {
    let directory = Cc["@mozilla.org/file/directory_service;1"]
                      .getService(Ci.nsIProperties)
                      .get("ProfD", Ci.nsIFile);
    directory.append("extensions");
    directory.append(EXTENSION_UUID);
    directory.append("db");

    /* make sure the directory exists */
    if (!directory.exists()) {
      directory.create(Ci.nsIFile.DIRECTORY_TYPE, 0755);
    }
     
 
    let file = directory.clone(); 
    file.append("config.db");

    if (!(storageService instanceof Ci.mozIStorageService)) {
      storageService = Cc["@mozilla.org/storage/service;1"]
                          .getService(Ci.mozIStorageService);
    }

    let dbConn = storageService.openDatabase(file); 
                    // Will also create the file if it does not exist
                    
    if (!(dbConn instanceof Ci.mozIStorageConnection)
        || !dbConn.connectionReady) {
      throw Components.results.NS_ERROR_FAILURE;
    }

    const createQuery = "CREATE TABLE IF NOT EXISTS \"main\".\"libraries\" ( "
                            +"\"id\" VARCHAR PRIMARY KEY  NOT NULL , "
                            +"\"name\" VARCHAR NOT NULL , "
                            +"\"address\" VARCHAR NOT NULL , "
                            +"\"port\" INTEGER NOT NULL "
                            +");";

    dbConn.executeSimpleSQL(createQuery);

    return dbConn;

  }, //openConfigDatabase


  /* deleteFromConfig
   *  Remnve library paramaters from persistent storage.
   *
   * \param aDatabaseId
   *
   **/

  deleteFromConfig: function SbDaapLibManager_deleteFromConfig (aDatabaseId)
  {
    /* preconditions */
    if (!aDatabaseId || !aDatabaseId.length) {
      /* abort */
      throw Components.results.NS_ERROR_INVALID_ARG; 
    }

    let dbConn = this.openConfigDatabase();

    let deleteQuery = "DELETE FROM \"main\".\"libraries\" "
                        +"WHERE \"id\" = \""+aDatabaseId+"\";";

    dbConn.executeSimpleSQL(deleteQuery);
    dbConn.close();

    return ;
  },  //deleteFromConfig

}; //SbDaapLibManager


/*=============================================================================*/
/* SbDaapLibManager Commands Callbacks                                         */
/*=============================================================================*/

/* cmdLogoutExecute:
 *  Close this library and logout from the server.
 *
 *  Implements: sbIPlaylistCommandsBuilderSimpleCallback
 *
 *  \param aContext
 *  \param aSubMenuId
 *  \param aCommandId
 *  \param aHost
 *
 **/
function SbDaapLibManager_cmdLogoutExecute 
                    ( aContext, aSubMenuId, aCommandId, aHost) 
{
  /* get the library */
  let databaseId = aContext.medialist.library
                           .getProperty(DProperty.databaseId);

  // XXXDaniel: don't use "this" because it's a static function 
  if (!(daapLibManager instanceof Ci.sbIDaapLibraryManager)) {
    daapLibManager = Cc[CONTRACT_ID].getService(Ci.sbIDaapLibraryManager);
  }

  daapLibManager.close(databaseId);

  //XXXDaniel: what now? go to the main library? go back? I don't know...
  var gBrowser =  Cc["@mozilla.org/appshell/window-mediator;1"]
                   .getService(Ci.nsIWindowMediator)
                   .getMostRecentWindow("Songbird:Main").window.gBrowser;
  //gBrowser.selectedBrowser.webNavigation.goBack();
  gBrowser.loadMediaList(libraryManager.mainLibrary);

}


/* cmdLogoutVisible:
 *  Display the logout command only when is appropiated
 *
 *  Implements: sbIPlaylistCommandsBuilderSimpleCallback
 *
 *  \param aContext
 *  \param aSubMenuId
 *  \param aCommandId
 *  \param aHost
 *
 **/
function SbDaapLibManager_cmdLogoutVisible
                    ( aContext, aSubMenuId, aCommandId, aHost) 
{
  /* get the library */
  let databaseId = aContext.medialist.library
                           .getProperty(DProperty.databaseId);

  // XXXDaniel: don't use "this" because it's a static function 
  if (!(daapLibManager instanceof Ci.sbIDaapLibraryManager)) {
    daapLibManager = Cc[CONTRACT_ID].getService(Ci.sbIDaapLibraryManager);
  }

  return daapLibManager.isOpen(databaseId);
}


/* cmdForceUpdateExecute: 
 *  Force a library reload.
 *
 *  Implements: sbIPlaylistCommandsBuilderSimpleCallback
 *
 *  \param aContext
 *  \param aSubMenuId
 *  \param aCommandId
 *  \param aHost
 *
 **/
function SbDaapLibManager_cmdForceUpdateExecute
                    ( aContext, aSubMenuId, aCommandId, aHost) 
{
  /* get the library */
  let library = aContext.medialist.library;
  let databaseId = library.getProperty(DProperty.databaseId);


  // XXXDaniel: don't use "this" because it's a static function 
  if (!(daapLibManager instanceof Ci.sbIDaapLibraryManager)) {
    daapLibManager = Cc[CONTRACT_ID].getService(Ci.sbIDaapLibraryManager);
  }

  /* schedule an update on the next access */
  daapLibManager.forceUpdate(databaseId);

  /* reload the library to actually do the update */
  if (typeof(window) == "undefined") {
    var window =  Cc["@mozilla.org/appshell/window-mediator;1"]
                   .getService(Ci.nsIWindowMediator)
                   .getMostRecentWindow("Songbird:Main").window;
  }
  window.setTimeout(function() { window.gBrowser.loadMediaList(library); }, 0);

} //cmdForceUpdateExecute


/* cmdForceUpdateVisible:
 *  Display the force update command only when is appropiated
 *
 *  Implements: sbIPlaylistCommandsBuilderSimpleCallback
 *
 *  \param aContext
 *  \param aSubMenuId
 *  \param aCommandId
 *  \param aHost
 *
 **/
function SbDaapLibManager_cmdForceUpdateVisible
                    ( aContext, aSubMenuId, aCommandId, aHost) 
{
  /* get the library */
  let databaseId = aContext.medialist.library
                           .getProperty(DProperty.databaseId);

  // XXXDaniel: don't use "this" because it's a static function 
  if (!(daapLibManager instanceof Ci.sbIDaapLibraryManager)) {
    daapLibManager = Cc[CONTRACT_ID].getService(Ci.sbIDaapLibraryManager);
  }

  return daapLibManager.isOpen(databaseId);
}



/* cmdDeleteExecute:
 *  Delete a static library
 *
 *  Implements: sbIPlaylistCommandsBuilderSimpleCallback
 *
 *  \param aContext
 *  \param aSubMenuId
 *  \param aCommandId
 *  \param aHost
 *
 **/
function SbDaapLibManager_cmdDeleteExecute
                    ( aContext, aSubMenuId, aCommandId, aHost) 
{
  /* get the library */
  let library = aContext.medialist.library;
  let databaseId = library.getProperty(DProperty.databaseId);


  // XXXDaniel: don't use "this" because it's a static function 
  if (!(daapLibManager instanceof Ci.sbIDaapLibraryManager)) {
    daapLibManager = Cc[CONTRACT_ID].getService(Ci.sbIDaapLibraryManager);
  }

  /* delegate the delete to the library manager */
  //daapLibManager.delete(databaseId);

  /* reload the library to actually do the update */
  if (typeof(window) == "undefined") {
    var window =  Cc["@mozilla.org/appshell/window-mediator;1"]
                   .getService(Ci.nsIWindowMediator)
                   .getMostRecentWindow("Songbird:Main").window;
  }
  window.setTimeout(function() { window.gBrowser.loadMediaList(libraryManager.mainLibrary); }, 0);
  window.setTimeout(function() { daapLibManager.delete(databaseId); }, 1000);

} //cmdDeleteExecute


/* cmdDeleteVisible:
 *  Display the delete library command only when is appropiated
 *
 *  Implements: sbIPlaylistCommandsBuilderSimpleCallback
 *
 *  \param aContext
 *  \param aSubMenuId
 *  \param aCommandId
 *  \param aHost
 *
 **/
function SbDaapLibManager_cmdDeleteVisible
                    ( aContext, aSubMenuId, aCommandId, aHost) 
{
  /* get the library */
  let databaseId = aContext.medialist.library
                           .getProperty(DProperty.databaseId);

  // XXXDaniel: don't use "this" because it's a static function 
  if (!(daapLibManager instanceof Ci.sbIDaapLibraryManager)) {
    daapLibManager = Cc[CONTRACT_ID].getService(Ci.sbIDaapLibraryManager);
  }

  return daapLibManager.isStatic(databaseId);
} //cmdDeleteVisible



/*=============================================================================*/
/* DaapLibrary Class                                                           */
/*=============================================================================*/

/* constructor */
function DaapLibrary ()
{
  //modules
  if (typeof(DProperty) == "undefined") {
    Components.utils.import("resource://daapclient/DaapProperties.jsm");
  }
  if (typeof(DaapImporterJob) == "undefined") {
    Components.utils.import("resource://daapclient/DaapImporterJob.jsm");
  }
  if (typeof(SBProperties) == "undefined") {
    Components.utils.import("resource://app/jsmodules/sbProperties.jsm");
  }
  if (typeof(LibraryUtils) == "undefined") {
    Components.utils.import("resource://app/jsmodules/sbLibraryUtils.jsm");
  }

  //global variables
  if (!(libraryManager instanceof(Ci.sbILibraryManager))) {
    libraryManager = Cc["@songbirdnest.com/Songbird/library/Manager;1"]
                      .getService(Ci.sbILibraryManager);
  }
  if (!(librarySPS instanceof(Ci.sbILibraryManager))) {
    librarySPS = Cc['@songbirdnest.com/servicepane/library;1']
                  .getService(Ci.sbILibraryServicePaneService);
  }
  if (!(ioService instanceof(Ci.nsIIOService))) {
    ioService = Cc["@mozilla.org/network/io-service;1"]
                 .getService(Ci.nsIIOService);
  }
  if (!(l10n instanceof Ci.nsIStringBundleService)) {
    l10n = Cc["@mozilla.org/intl/stringbundle;1"]
            .getService(Ci.nsIStringBundleService)
            .createBundle(PROPERTY_FILE);
  }
  if (!(SPS instanceof Ci.sbIServicePaneService)) {
    SPS = Cc['@songbirdnest.com/servicepane/service;1']
           .getService(Ci.sbIServicePaneService);
    SPS.init();
  }
  if (!(BMS instanceof Ci.sbIBookmarks)) {
    BMS = Cc['@songbirdnest.com/servicepane/bookmarks;1']
               .getService(Ci.sbIBookmarks);
  }


  //member variables
  this._library = null;
  this._playLists = null;
  this._sync = null;

  this._databaseId = null;
  this._machineName = null;
  this._address = null;
  this._port = null;
  this._uri = null;

  this._libraryNode = null;
  this._client = null;
  this._database = null;

  /* set the status */
  this._open = false;
  this._forceUpdate = false;
}


DaapLibrary.prototype = {

  /* private data ---------------------------------------------------------*/
  _databaseId: null,
  _machineName: null,
  _address: null,
  _port: null,
  _uri: null,

  _library: null,
  _playLists: null,
  _open: null,
  _forceUpdate: null,

  _libraryNode: null,
  _client: null,
  _database: null,

  /* public functions -----------------------------------------------------*/


  /* load:
   *  Initialize a new library and shows the node in the service pane.
   *
   * \param aDatabaseId
   * \param aMachineName
   * \param aAddress
   * \param aPort
   *
   **/

  load: function DaapLibrary_load (aDatabaseId, aMachineName, aAddress, aPort)
  {
    /* preconditions */
    if (!aDatabaseId || !aDatabaseId.length
        || !aMachineName || !aMachineName.length
        || !aAddress || !aAddress.length
        || !aPort ) {
      /* abort */
      throw Components.results.NS_ERROR_INVALID_ARG; 
    }

    /* save */
    this._databaseId = aDatabaseId;
    this._machineName = aMachineName;
    this._address = aAddress;
    this._port = aPort;

    let url = "daap://" + this._address + ":" + this._port + "/";
    this._uri = ioService.newURI(url, null, null);
    if (!this._uri) {
      /* abort */
      throw Cr.NS_ERROR_FAILURE;
    }

    /* get related library */
    this._library = this.loadLibrary(aDatabaseId);
    if (!this._library) {
      /* cannot load the library */
      throw Cr.NS_ERROR_FAILURE;
    }

    /* set the current name */
    this._library.name = aMachineName;

    ///* hide all the playlists */
    //this.hidePlayLists();

    /* register the library */
    libraryManager.registerLibrary(this._library, false);

    /* hide all the playlists */
    this.hidePlayLists();

    /* add service pane node */
    this._libraryNode = this.addNode(this._library);

  },  //load



  /* unload:
   *  Sort of destructor. Called when a daap library isn't available anymore.
   *
   **/
  unload: function DaapLibrary_unload ()
  {
    /* preconditions */
    if (!(this._library instanceof Ci.sbILibrary)) {
      /* abort */
      return;
    }

    /* hide all the playlists */
    try {
      //XXX Daniel: when unregistering a library only childs are removed
      //  but if there are playlists node on the same level they persist there
      //  is it a bug? XXX
      this.hidePlayLists();
    }
    catch (e) {}

    /* unregister the library */
    try {
      libraryManager.unregisterLibrary(this._library);
    }
    catch (e) {
      Components.utils.reportError(e);
    }

    /* delete the node */
    //Note: when unregistering the library the related node is hidden not deleted
    if (this._libraryNode) {
      try {
        let node = this._libraryNode.parentNode;
        node.removeChild(this._libraryNode);
      }
      catch (e) {}
      finally { 
        this._libraryNode = null;
      }
    }
    
    /* clean up */
    this._library = null;
    this._libraryNode = null;

    this._databaseId = null;
    this._machineName = null;
    this._address = null;
    this._port = null;
    this._uri = null;

    return;
  },  //unload



  /* open:
   *  Open/use the current library, connect to the server, 
   *  download information, create playlists, sync tracks, etc.
   *
   **/

  open: function DaapLibrary_open ()
  {
    /* preconditions */
    if (!(this._library instanceof Ci.sbILibrary) 
        || !(this._uri instanceof Ci.nsIURI)) {
      /* abort */
      throw Components.results.NS_ERROR_FAILURE;

    }

    /* speed shortcut */
    if (this.isOpen()) {
      /* already opened */
      return;
    }

    /* check if an update has been requested */
    if (this._forceUpdate) {
      this._forceUpdate = null;
      this._library.clear();
    }

    /* create the client */
    this._client = Cc["@songbirdnest.com/daap/client;1"]
                      .createInstance(Ci.sbIDaapClient);
    if (!this._client) {
      /* abort */
      throw Components.results.NS_ERROR_OUT_OF_MEMORY;
    }

    try {
      this._client.init(this._uri);
    }
    catch (e if e.result == Components.results.NS_ERROR_NOT_AVAILABLE) {
      /* notify the user */
      this.warnConnectionError(this._uri.spec);

      //TODO: review this...
      //XXXcopied from the login fail below:
      if (typeof(window) == "undefined") {
        var window =  Cc["@mozilla.org/appshell/window-mediator;1"]
                       .getService(Ci.nsIWindowMediator)
                       .getMostRecentWindow("Songbird:Main").window;
      }
      window.setTimeout(function() { 
              window.gBrowser.loadMediaList(libraryManager.mainLibrary); }, 0);


      /* undo until here */
      this._client = null;


      /* log the error */
      Components.utils.reportError("[Daap Client] cannot connect to "+this._uri.spec);
      return;
    }

    /* get server info */
    let serverInfo = this._client.getServerInfo();
    //this.supportUpdate = serverInfo.supportUpdate;
    let protocol = parseFloat(serverInfo.protocol);
    if (protocol > 3.2) {
      /* protocol unsupported, probably iTunes 7 or iTunes 8 */
      this.warnProtocol();
      return;
    }

    /* login */
    try {
      this._client.login();
    }
    catch (e) {
      /* login fails, abort... */
      //XXXDaniel: empty the library so we don't show unavailable tracks.
      //this._library.clear();
      //XXXDaniel: go to the main library?
      if (typeof(window) == "undefined") {
        var window =  Cc["@mozilla.org/appshell/window-mediator;1"]
                       .getService(Ci.nsIWindowMediator)
                       .getMostRecentWindow("Songbird:Main").window;
      }
      window.setTimeout(function() { 
              window.gBrowser.loadMediaList(libraryManager.mainLibrary); }, 0);


      /* undo until here */
      this._client = null;

      /* notify the client */
      Components.utils.reportError("[Daap Client] login failed!");

      return;
    }

    /* get database info */
    let databases = this._client.getDatabases();
    if (!databases || !databases.length) {
      /* abort */
      return;
    }
   
    // XXX: always use the first database (the library!)
    this._database = databases.queryElementAt(0, Ci.sbIDaapDatabase);

    /* save the current database id */
    this._library.setProperty(DProperty.id, this._database.id);


    // Don't set name if it hasn't changed to avoid an UI redraw
    if (this._library.name != this._database.name) {
      this._library.name = this._database.name;
    }

    //XXXDaniel don't uncomment this because the node is unselected when renamed    
    //if (this._libraryNode.name =! this._database.name) {
    //  this._libraryNode.name = this._database.name;
    //}

    /* decide here if we'll update or not so we can clear the library
     *  before creating the playlists 
     ***/
    let sync = this.isSync();
    if (!sync) {
      /* empty the library */
      this._library.clear();
    }

    /* create playlists */
    this.createPlaylists();

    /* delete older playlists */
    this.cleanMediaLists();

    /* display the playlists */
    for (let i in this._playLists) {
      //XXXDaniel don't set the property unless is necesary because is slow
      /*
      let hidden = this._playLists[i].mediaList.getProperty(SBProperties.hidden);
      if (hidden == "1") {
        this._playLists[i].mediaList.setProperty(SBProperties.hidden, "0");
      }
      */
      this._playLists[i].node.hidden = false;
    }


    /* update the library */
    if (!sync) {
      this._job = new DaapImporterJob();
      this._job.setTitleText(l10n.GetStringFromName("DlgImportTitle"));
      this._job.setStatusText(l10n.GetStringFromName("DlgImportMessage"));
      this._job.start(this._client, this._library, this._database.itemCount);
    }
    else {
      /* remember for the next time */
      this._open = true;
    }

    return;
  },  //open


  /* openPlaylist:
   *  Open/use a playlist, connect to the server and sync tracks if appropiated.
   *
   *  \param aMediaList   the playlist
   *
   **/

  openPlaylist: function (aMediaList)
  {
    /* validate arguments */
    if (!(aMediaList instanceof Ci.sbIMediaList)) {
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    /* make sure the library is open */
    if (!this.isOpen()) {
      //hide the playlists nodes
      this.hidePlayLists();
      //XXXDaniel: what now? go to the main library? go back? I don't know...
      var gBrowser =  Cc["@mozilla.org/appshell/window-mediator;1"]
                        .getService(Ci.nsIWindowMediator)
                        .getMostRecentWindow("Songbird:Main").window.gBrowser;
      gBrowser.loadMediaList(libraryManager.mainLibrary);
      //gBrowser.selectedBrowser.webNavigation.goBack();
      return;
    }

    /* preconditions */
    if (!this._library
        || !this._client
        || !this._playLists) {
      /* abort */
      throw Components.results.NS_ERROR_FAILURE;
    }

    /* validate the playlist */
    let id = aMediaList.getProperty(DProperty.id);
    if (!id || !this._playLists[id]) {
      /* invalid playlist */
      return;
    }

    if (this.isOpen(this._playLists[id])) {
      /* already opened */
      return;
    }

    /* check if we need to sync */
    if (!this.isSync(this._playLists[id].mediaList, this._playLists[id])) {
      this._playLists[id].mediaList.clear();
      this._playLists[id]._job = new DaapImporterJob();
      this._playLists[id]._job.setTitleText(l10n.GetStringFromName("DlgImportTitle"));
      this._playLists[id]._job.setStatusText(l10n.GetStringFromName("DlgImportMessage"));
      this._playLists[id]._job.start(this._client, 
                                      this._playLists[id].mediaList, 
                                      this._playLists[id].itemCount);
    }
    else {
      /* remember for the next time */
      this._playLists[id]._open = true;
    }

    return;
  },  //openPlaylist



  /* close:
   *  Close the library and logout from the server.
   *  (reverse whatever is what open does!) 
   *
   **/

  close: function DaapLibrary_close ()
  {
    if (!this.isOpen()) {
      /* nothing to do */
      return;
    }

    /* logout from the server */
    if (this._client) {
      this._client.logout();
    }
    this._client = null;

    /* hide the playlists */
    this.hidePlayLists();

    /* clean up */
    this._database = null;
    this._playLists = null;
    this._job = null;  

    /* update the status */
    this._open = false;

    return;
  }, //close



  /* isOpen:
   *  Determine if the library (or play list) is open.
   *
   * \param aList [optional] a play list to check, if ommited
   *                the library is checked.
   *
   * \return boolean ture if it's open false if it isn't
   *
   **/

  isOpen: function DaapLibrary_isOpen (aList)
  {
    if (typeof(aList) == "undefined") {
      aList = this;
    }
    if (aList._open) {
      return true;
    }
    if (this.checkJob(aList._job)) {
      return true;
    }
    return false;
  }, //isOpen



  /* getLibrary:
   *  Return the sbILibrary object associated with this library.
   *
   *  \return sbILibrary  a library.
   *
   **/

  getLibrary: function DaapLibrary_getLibrary ()
  {
    if (!this._library) {
      throw Components.results.NS_ERROR_FAILURE;
    }
      
    return this._library;
  },  //getLibrary



  /* forceUpdate:
   *  Setup and update the next time the library is opened.
   *
   **/
  forceUpdate: function DaapLibrary_forceUpdate ()
  { 
    /* set the update flag */
    this._forceUpdate = true;

    /* close the library */
    //XXX maybe this has to be done elsewhere
    this.close();

  }, //forceUpdate



  /* delete:
   *  Delete the library file from it physical location
   *
   **/
/*
  delete: function DaapLibrary_delete ()
  {
    if (!this._library) {
      throw Components.results.NS_ERROR_FAILURE;
    }

    if (this.isOpen()) {
      this.close();
    }

    try {
      let file = this._library.creationParameters
                     .getPropertyAsInterface("databaseFile", Ci.nsIFile);
      file.remove (false); //false stands for not recursive, ignored for files
    }
    catch (e) {
      Components.utils.reportError(e);
    }

    return;
  }, //delete
*/




  /* private functions ----------------------------------------------------*/


  /* createLibrary:
   *  Create a new local library using for the databaseId,
   *
   *  \param  aDatabaseId  the database Id
   *  \return library     sbILibrary object or null
   *
   **/

  createLibrary: function DaapLibrary_createLibrary (aDatabaseId) 
  {
    if (!aDatabaseId || !aDatabaseId.length) {
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    let directory = Cc["@mozilla.org/file/directory_service;1"]
                      .getService(Ci.nsIProperties)
                      .get("ProfD", Ci.nsIFile);
    directory.append("extensions");
    directory.append(EXTENSION_UUID);
    directory.append("db");
     
    let file = directory.clone(); 
    file.append("daap_" +aDatabaseId + ".db");

    let libraryFactory =
      Cc["@songbirdnest.com/Songbird/Library/LocalDatabase/LibraryFactory;1"]
        .getService(Ci.sbILibraryFactory);

    let hashBag = Cc["@mozilla.org/hash-property-bag;1"]
                    .createInstance(Ci.nsIWritablePropertyBag2);
    hashBag.setPropertyAsInterface("databaseFile", file);

    let library = libraryFactory.createLibrary(hashBag);

    return library;
  },  //createLibrary



  /* createPlaylists: 
   *  Replicate the daap playlists on this library,
   *  deleting any other foreign media list and 
   *  creating the associated nodes.
   *
   **/

  createPlaylists: function DaapLibrary_createPlaylists ()
  {
    if (!(this._library instanceof Ci.sbILibrary)
        || !(this._client instanceof Ci.sbIDaapClient)
        || !(this._libraryNode instanceof Ci.sbIServicePaneNode)
        || !this._database ) {
      /* abort */
      throw Components.results.NS_ERROR_FAILURE;
    }

    /* initialize */
    this._playLists = {};


    /* download containers */
    let containers = this._client.getPlaylists(this._database.id);
    if (!containers) {  
      throw Components.results.NS_ERROR_FAILURE;
    }


    /* add playlists */
    let length = containers.length;
    for (let i=1; i<length; i++) {
      let container = containers.queryElementAt(i, Ci.sbIDaapPlaylist);
      let mediaList = this.getMediaList(container.id, container.name);
      let node = this.addNode(mediaList, this._libraryNode);

      if (!(mediaList instanceof Ci.sbIMediaList) 
          || !(node instanceof Ci.sbIServicePaneNode)) {
        /* abort */
        throw Components.results.NS_ERROR_FAILURE;
      }

      this._playLists[container.id] = {
          id: container.id,
          name: container.name,
          itemCount: container.itemCount,
          mediaList: mediaList,
          node: node,
          _open: false,
          _job: null
        };

    }//for

    return;
  },  //createPlaylists


  /* cleanMediaLists:
   *  Delete all media lists that don't match a daap playlist.
   *
   **/
  cleanMediaLists: function DaapLibrary_cleanMediaLists ()
  {
    if (!(this._library instanceof Ci.sbILibrary)
        || !this._playLists) {
      throw Components.results.NS_ERROR_FAILURE;
    }


    let lists = null;
    try {
      /* get the media lists */
      lists = this._library.getItemsByProperty(SBProperties.isList, "1");
    }
    catch (e if e.result == Components.results.NS_ERROR_NOT_AVAILABLE) {
      /* no playlists */
    }

    if (!lists || !lists.length) {
      /* no playlists */
      return;
    }

    let length = lists.length;
    for (let i=0; i<length; i++) {
      let mediaList = lists.queryElementAt(i, Ci.sbIMediaList);
      if (mediaList) {
        let id = null;
        try {
          id = mediaList.getProperty(DProperty.id);
        }
        catch (e) {};

        if (!id || !this._playLists[id]) {
          /* delete it */
          this.removePlayList(mediaList);
        }
      }
    }

  }, //cleanMediaLists



  /* getMediaList:
   *  Get or create a media list associated with the daap container Id.
   *
   * \param aId the daap container Id
   * \param aName [optional] the new playlist name
   *
   * \return sbIMediaList a media list
   *
   **/

  getMediaList: function DaapLibrary_getMediaList (aId, aName)
  {
    /* preconditions */
    if (!aId) {
      throw Components.results.NS_ERROR_INVALID_ARG;
    }
    if (!(this._library instanceof Ci.sbILibrary)) {
      throw Components.results.NS_ERROR_FAILURE;
    }


    /* search */
    try {
      /* get the media lists */
      lists = this._library.getItemsByProperty(DProperty.id, aId);
    }
    catch (e if e.result == Components.results.NS_ERROR_NOT_AVAILABLE) {
      /* no media lists */
      lists = null;
    }

    let mediaList = null;

    if (lists && lists.length > 0) {
      /* there at least one, use it */
      mediaList = lists.queryElementAt(0, Ci.sbIMediaItem);

      /* delete the others */
      let length = lists.length;
      for (let i=1; i<length; i++) {
        try {
          this.removePlayList(lists.queryElementAt(i, Ci.sbIMediaItem));
        }
        catch (e) {}
      }
    }


    if (!(mediaList instanceof Ci.sbIMediaList)) {
      /* try to create a new one */

      let properties = 
            Cc["@songbirdnest.com/Songbird/Properties/MutablePropertyArray;1"]
             .createInstance(Ci.sbIMutablePropertyArray);
      properties.appendProperty(DProperty.id, aId);
      //properties.appendProperty(SBProperties.hidden, "1");
      properties.appendProperty(SBProperties.isReadOnly, "1");
      mediaList = this._library.createMediaList("simple", properties);  

    }

    
    if (!(mediaList instanceof Ci.sbIMediaList)) {
      /* ooops */
      throw Components.results.NS_ERROR_FAILURE;
    }


    /* set properties */
    //XXXDaniel: this is very slow!!
    //            don't do it unles very necesary
    //mediaList.setProperty(DProperty.id, aId);
    //mediaList.setProperty(SBProperties.hidden, "1");
    //mediaList.setProperty(SBProperties.isReadOnly, "1");
    

    if (aName) {
      mediaList.name = aName;
    }
    
    return mediaList;
  }, //getMediaList



  /* loadLibrary:
   *  Search a local library that matchs the databaseId,
   *  create one if it isn't found.
   *
   *  \param  aDatabaseId  the database Id
   *  \return library     sbILibrary object or null
   *
   **/

  loadLibrary: function DaapLibrary_loadLibrary (aDatabaseId)
  {
    
    if (!aDatabaseId || !aDatabaseId.length) {
      throw Components.results.NS_ERROR_INVALID_ARG;
    }
    
    let enumerator = libraryManager.getLibraries();
    while (enumerator.hasMoreElements()) {
      let library = enumerator.getNext().QueryInterface(Ci.sbILibrary);
      if (library && library.getProperty(DProperty.databaseId) == aDatabaseId) {
        return library;
      }
    }

    let library = this.createLibrary(aDatabaseId);
    if (library) {
      library.setProperty(DProperty.databaseId, aDatabaseId);
      library.setProperty(SBProperties.isReadOnly, "1");
      return library;
    }

    return null;
  },  //loadLibrary



  /* addNode:
   *  Create the service pane nodes for the libraries or the playlists.
   *
   *  \param aMediaList  a library or playlist
   *  \param aNode       [optional] the parent node, is empty the new node 
   *                     is created as a child of the "shared" folder.
   *  \return node       the new node
   *
   **/

  addNode: function DaapLibrary_addNode (aMediaList, aNode)
  {

    let node = librarySPS.getNodeForLibraryResource(aMediaList);
    if (!node) {
      /* cannot create the node, abort */
      return null;
    }

    let root = null;
    if (aNode instanceof Ci.sbIServicePaneNode) {
      /* use the given node */
      root = aNode;
    }
    else {
      /* get the folder node */
      root = this.getFolderNode();
    }

    if (root) {
      //XXXDaniel: don't append the node if it isn't needed (for speed up)
      if (!(node.parent instanceof Ci.sbIServicePaneNode)
          || !(node.parent.id == root.id) ){
        root.appendChild(node);
      }
    }

    /* set node properties */
    if (node.name != aMediaList.name) {
      node.name = aMediaList.name;
    }
    if (node.editable) {
      node.editable = false;
    }
    if (node.hidden) {
      node.hidden = false;
    }

    //XXXDaniel: do not unhide only the node, unhide the medialist too
    //XXXDaniel don't set the property unless is necesary because is slow
    /*
    let hidden = aMediaList.getProperty(SBProperties.hidden);
    if (hidden == "1") {
      aMediaList.setProperty(SBProperties.hidden, "0");
    }
    */

    return node;
  },  //addNode



  /* getFolderNode:
   *  Get or create the shared folder node.
   *
   *  \return sbIServicePaneNode  folder node
   *
   **/

  getFolderNode: function DaapLibrary_getFolderNode ()
  {
    let node = this.findFolderNode();
    if (node == null) {
      /* compatibility with previous daapclient */
      node = this.findNode(l10n.GetStringFromName("sharedFolderLabel"));
    }
    if (node == null) {
      /* create the folder if it isn't found */
      node = BMS.addFolder(l10n.GetStringFromName("sharedFolderLabel")); 
    }

    /* Sort the remoteLibFolder node in the service pane */
    const SP = "http://songbirdnest.com/rdf/servicepane#";
    node.setAttributeNS(SP, "Weight", 1);
    node.setAttributeNS("daap", "sharedFolder", true);

    SPS.sortNode(node)
      
    node.editable = true;

    return node;
  },  //getFolderNode


  /* findNode: 
   *  Walk through the service pane tree searching for a node by name.
   *
   *  \param  nodeName   searched node name
   *  \param  startNode  [optional] root node for this search
   *  \return sbIServicePaneNode if found or null
   *
   **/

  findNode: function DaapLibrary_findNode (nodeName, startNode) 
  {
    if (!startNode) {
      startNode = SPS.root;
    }
    
    /* startNode must be a container or we're done */
    if (!startNode.isContainer) {
      return null;
    }

    /* check if this is our node */
    if(startNode.name == nodeName) {
      return startNode;
    }

    /* search between our children */
    let children = startNode.childNodes;
    while (children.hasMoreElements()) {
      let child = children.getNext().QueryInterface(Ci.sbIServicePaneNode);
      let result = this.findNode(nodeName, child);
      if (result != null) {
        return result;
      }
    }
  
    /* not found! */
    return null;
  },  //findNode



  /* findFolderNode
   *  Walk through the service pane tree searching for the shared
   *  libraries folder.
   *
   *  \param  aNode  [optional] root node for this search
   *  \return sbIServicePaneNode if found or null
   *
   **/

  findFolderNode: function DaapLibrary_findFolderNode (aNode) 
  {
    if (!aNode) {
      aNode = SPS.root;
    }
    
    /* aNode must be a container or we're done */
    if (!aNode.isContainer) {
      return null;
    }

    /* check if this is our node */
    let found = aNode.hasAttributeNS("daap", "sharedFolder");
    if(found) {
      return aNode;
    }

    /* search between our children */
    let children = aNode.childNodes;
    while (children.hasMoreElements()) {
      let child = children.getNext().QueryInterface(Ci.sbIServicePaneNode);
      let result = this.findFolderNode(child);
      if (result != null) {
        return result;
      }
    }
  
    /* not found! */
    return null;
  },  //findFolderNode



  /* warnProtocol:
   *   Display a message indicating the server has an unsupported protocol
   *
   **/

  warnProtocol: function DaapLibrary_warnProtocol ()
  {
    /* get the main window */
    if (typeof(window) == "undefined") { 
      var window = Cc["@mozilla.org/appshell/window-mediator;1"]
                    .getService(Ci.nsIWindowMediator)
                    .getMostRecentWindow("Songbird:Main").window;
    }

    /* prompt service */
    let prompts = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                    .getService(Ci.nsIPromptService);

    /* show the message */
    prompts.alert(window, 
                  l10n.GetStringFromName("DlgErrorTitle"), 
                  l10n.GetStringFromName("ErrorUnsupportedProtocol"));
  
    return;
  },  //warnProtocol



  /* warnConnectionError:
   *   Display a message indicating a connection error
   *
   * \param aURL  the server url
   *
   **/

  warnConnectionError: function DaapLibrary_warnConnectionError (aURL)
  {
    /* get the main window */
    if (typeof(window) == "undefined") { 
      var window = Cc["@mozilla.org/appshell/window-mediator;1"]
                    .getService(Ci.nsIWindowMediator)
                    .getMostRecentWindow("Songbird:Main").window;
    }

    /* prompt service */
    let prompts = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                    .getService(Ci.nsIPromptService);

    /* show the message */
    prompts.alert(window, 
                  l10n.GetStringFromName("DlgErrorTitle"), 
                  l10n.formatStringFromName("ErrorConnectionError", [aURL], 1));
  
    return;
  },  //warnConnectionError



  /* hidePlayLists:
   *  Hide all the media lists in the current library
   *  and they nodes.
   *
   **/

  hidePlayLists: function DaapLibrary_hidePlayLists ()
  {
    /* preconditions */
    if (!(this._library instanceof Ci.sbILibrary)) {
      /* abort */
      throw Components.results.NS_ERROR_FAILURE;
    }


    let lists = null;
    try {
      /* get the media lists */
      lists = this._library.getItemsByProperty(SBProperties.isList, "1");
    }
    catch (e) {
      /* no playlists */
      return;
    }


    let length = lists.length;

    if (!length) {
      /* no playlists */
      return;
    }

    /* traverse the play lists */
    for (let i=0; i<length; i++) {
      let mediaItem = lists.queryElementAt(i, Ci.sbIMediaItem);
      if (mediaItem) {
        //mediaItem.setProperty(SBProperties.hidden, "1");
        let node = librarySPS.getNodeForLibraryResource(mediaItem);
        if (node && !node.hidden) {
          node.hidden = true;
        }
      }//if
    }//for

    return;
  },  //hidePlayLists


  
  /* removePlayList:
   *  Remove the given mediaItem from the current library
   *  and his associated node from the service pane if one 
   *  exists.
   *
   *  \param  aMediaItem  the item to remove
   *
   **/
  removePlayList: function DaapLibrary_removePlayList (aMediaItem)
  {
    /* preconditions */
    if (!(this.library instanceof(Ci.sbILibrary))
        || !(aMediaItem instanceof(Ci.sbIMediaItem))) {
      /* abort */
      return;
    }

    /* try to remove the node */
    let node = librarySPS.getNodeForLibraryResource(aMediaItem);
    if (node) {
      node.parentNode.removeChild(node);
    }

    /* remove the mediaItem */
    this.library.remove(aMediaItem);
    
    return;
  },  //removePlayList



  /* isSync
   *  Decide if the library (or playlist) needs to be updated or not.
   *
   * \param aMediaList  [optional] the media list to check, defaults to the library
   * \param aContainer  [optional] the related daap container
   *  \return isSync  true or false
   *
   **/

  isSync: function DaapLibrary_isSync (aMediaList, aContainer)
  {
    if (typeof(aMediaList) == "undefined") {
      /* defaults to the library */
      aMediaList = this._library;
      aContainer = this._database;
    }
    else if (typeof(aContainer) == "undefined") {
      throw Components.results.NS_ERROR_INVALID_ARG;
    }

    /* count ONLY the songs */
    let sbCount = this.getLength(aMediaList);
    let daapCount = aContainer.itemCount;

    if (daapCount != sbCount) {
      //dump(" isSync: FALSE -- DDDDDDDDDDDDD DEBUG tr#:"+daapCount+" ==? "+sbCount+"\n");
      return false; 
    }

    return true;

  },  //isSync



  /* getLength:
   *  Get the number of songs in a medialist
   *
   *  \param aMediaList the medialist
   *  \return length    track count
   *
   **/
  getLength: function DaapLibrary_getLength (aMediaList)
  {
    if (aMediaList instanceof Ci.sbIMediaList) {
      let view = aMediaList.createView();
      view.filterConstraint = LibraryUtils.standardFilterConstraint;
      return view.length
    }
    return null;
  },  //getLength



  /* checkJob:
   *  Check if the job is running or finished.
   *
   * \param aJob a DaapImporterJob instance
   *
   * \return bool true if the job is running or finished, 
   *                false anything else
   **/

  checkJob: function DaapLibrary_checkJob (aJob)
  {
    if (aJob instanceof DaapImporterJob) {
      if (aJob.status == Ci.sbIJobProgress.STATUS_SUCCEEDED)  {
        /* job finished succesfully */
        return true;
      }
      if (aJob.status == Ci.sbIJobProgress.STATUS_RUNNING) {
        /* job is runnning, return TRUE? */
        return true;
      }
      /* job error, try to cancel it */
      aJob.cancel();
    }

    /* no job, false */
    return false;
  }, //checkJob

};  //DaapLibrary



/**
 * Display debug messages and stats
 *
 **/

if (typeof(DaapDebug) == 'undefined') {

  function DaapDebug(aMessage, aStart) {

    if (aStart) {
      var elapsed = Date.now() - aStart;
      var time = (elapsed / 1000);
      aMessage += " (time: " + time + ")";
    }

    var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
                                 .getService(Components.interfaces.nsIConsoleService);
    consoleService.logStringMessage("[Daap Client] " + aMessage);
  }  
}



//=================================================
// Note: You probably don't want to edit anything
// below this unless you know what you're doing.
//
// Factory
var SbDaapLibManagerFactory = {
  singleton: null,
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    if (this.singleton == null)
      this.singleton = new SbDaapLibManager();
    return this.singleton.QueryInterface(aIID);
  }
};


/*=========================================================================*/
/* XPCOM Module                                                            */
/*=========================================================================*/

/* components constructors */
var components = [SbDaapLibManager];

/* NSGetModule entry point */
function NSGetModule(compMgr, fileSpec) {
  return XPCOMUtils.generateModule(components);
}

