/*
//
// BEGIN SONGBIRD GPL
//
// This file is part of the Songbird web player.
//
// Copyright(c) 2005-2008 POTI, Inc.
// http://songbirdnest.com
//
// This file may be licensed under the terms of of the
// GNU General Public License Version 2 (the "GPL").
//
// Software distributed under the License is distributed
// on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
// express or implied. See the GPL for the specific language
// governing rights and limitations.
//
// You should have received a copy of the GPL along with this
// program. If not, go to http://www.gnu.org/licenses/gpl.html
// or write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//
// END SONGBIRD GPL
//
*/

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cu = Components.utils;
const Cr = Components.results;

const CLASS_ID = Components.ID("{99a207c7-fc34-431a-b981-ecefd7a37fd2}");
const CLASS_NAME = "Daap Protocol Handler";
const CONTRACT_ID = "@mozilla.org/network/protocol;1?name=daap";

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");


/* global variables -------------------------------------------------------*/
var ioService = null;
var sessionManager = null;

/* sbDaapProtocolHandler constructor --------------------------------------*/
function sbDaapProtocolHandler() 
{
  if (!(ioService instanceof Ci.nsIIOService)) {
    ioService = Cc["@mozilla.org/network/io-service;1"]
                 .getService(Ci.nsIIOService);
  }
  if(!(sessionManager instanceof Ci.sbIDaapSessionManager)) {
    sessionManager = Cc["@songbirdnest.com/daap/session-manager;1"]
                      .getService(Ci.sbIDaapSessionManager);
  }

}

/* SbDaapProtocolHandler implementation -----------------------------------*/
sbDaapProtocolHandler.prototype = {
  /* nsISupports ---------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIProtocolHandler, 
                                          Ci.nsIClassInfo]),

  
  /* nsIClassInfo, so callers don't have to QI from JS -------------------*/
  classDescription        : CLASS_NAME,
  classID                 : CLASS_ID,
  contractID              : CONTRACT_ID,
  flags                   : Ci.nsIClassInfo.THREADSAFE,
  implementationLanguage  : Ci.nsIProgrammingLanguage.JAVASCRIPT,
  getHelperForLanguage    : function(aLanguage) { return null; },
  getInterfaces : function(count) {
    var interfaces = [Ci.nsIProtocolHandler, Ci.nsIClassInfo, Ci.nsISupports];
    count.value = interfaces.length;
    return interfaces;
  },

  /* nsIProtocolHandler ---------------------------------------------------*/

  /* readonly attribute ACString scheme; */
  scheme: "daap",
  /* readonly attribute long defaultPort; */
  defaultPort: 3689,
  /* readonly attribute unsigned long protocolFlags; */
  protocolFlags: Ci.nsIProtocolHandler.URI_STD 
                  | Ci.nsIProtocolHandler.ALLOWS_PROXY,


  /**
   * nsIURI newURI(in AUTF8String aSpec, 
   *               in string aOriginCharset, 
   *               in nsIURI aBaseURI); 
   */
  newURI: function(aSpec, aCharset, aBaseURI) 
  {
    let standardURL = Cc["@mozilla.org/network/standard-url;1"]
                         .createInstance(Ci.nsIStandardURL);

    if (standardURL) {
      standardURL.init(Ci.nsIStandardURL.URLTYPE_AUTHORITY, this.defaultPort, aSpec, aCharset, aBaseURI);
    }

    return standardURL;
  },

  /* nsIChannel newChannel(in nsIURI aURI); */
  newChannel: function(aURI)
  {

    /* check for an open session */
    let sessionId = null;
    try {
      sessionId = sessionManager.getSession(aURI);
    }
    catch (e) {
      //Cu.reportError(e);
    }

    if (sessionId) {
      /* add session id */
      let spec = aURI.spec;
      if (spec.search("session-id=") == -1) {
        let sep = spec.indexOf('?');
        if (sep >= 0) {
          sep += 1;
          spec = spec.substring(0, sep)
                    + "session-id=" + sessionId + "&"
                    + spec.substring(sep, spec.length);
        }
        else {
          spec += "?session-id=" + sessionId;
        }

        aURI.spec = spec;
      }
    }

    /* convert daap:// into http:// */
    queryURL = this.translateURI(aURI);

    let channel = ioService.newChannel(queryURL, null, null);

    /* replace the original URI */
    channel.originalURI = aURI;

    /* add standard request headers */
    let httpChannel = channel.QueryInterface(Ci.nsIHttpChannel);
    if (httpChannel) {
      httpChannel.setRequestHeader("Client-DAAP-Version", "3.0", false);
      httpChannel.setRequestHeader("Client-DAAP-Access-Index", "2", false);

      /* add session headers */
      try {
        httpChannel = sessionManager.addRequestHeaders(aURI, httpChannel);
      }
      catch (e) {
        //Cu.reportError(e);
      }

    }//if

    return channel;
  },

  /* boolean allowPort(in long port, in string scheme); */
  allowPort: function(port, scheme)
  {
    return true;
  },

  /* private methods ------------------------------------------------------*/

  /* given a DAAP URI translate it to a HTTP URL */
  translateURI: function (aURI)
  {
    /* validate the address */
    if (!aURI.schemeIs("daap") 
        || !aURI.host.length) {
      return null;
    }

    /* translate */
    let port = aURI.port;
    if (port == -1) {
      port = this.defaultPort;
    }

    let url = "http://" + aURI.host + (port!=80 ? ":"+port : "") + aURI.path;

    return url;
  }
}


/*=========================================================================*/
/* XPCOM Module                                                            */
/*=========================================================================*/

/* components constructors */
var components = [sbDaapProtocolHandler];

/* NSGetModule entry point */
function NSGetModule(compMgr, fileSpec) {
  return XPCOMUtils.generateModule(components);
}

